// infrared receiver for NEC code
// not using capture input, only pin change and interrupt

#include <stdint.h>
#include <avr/io.h> // for device register definitions
#include <avr/interrupt.h> // for interrupt handlers
#include "config.h"
#include "msg.h"
#include "profiler.h"
#include "ir_nec.h" // own interface


#ifdef CFG_PIN_IR
#define IR_PIN CFG_PIN_IR // pin definition
#else
#error Missing CFG_PIN_IR definition in config.h
#endif

#define TIMER0_PRESCALER 64 // define such that 8 bit count within timeout

// times in milliseconds
#define TIME_SHORT  0.56
#define TIME_LONG   1.69
#define TIME_AGC    9.0
#define TIME_SPACE  4.5
#define TIME_REPEAT 2.25
#define TIMEOUT    (TIME_AGC*TOLERANCE) // max. pulse length is 9 ms

#define TOLERANCE (1.3) // allow 30% deviation

#define MS2CLK(ms) ((uint8_t)((ms)*F_CPU/TIMER0_PRESCALER/1000))


/*
#define US2C(us) ((us)*F_CPU/TIMER0_PRESCALER/1000) // just for preprocessor, without cast

#if US2C(TIMEOUT) > 255
#error Prescaler too low
#endif

#if (US2C(TIMEOUT) < 127)
#warning Prescaler could be less
#endif
*/

static uint8_t nec_lastpin; // pin state of last round
static uint8_t nec_lasttime; // timer value of last round
static uint8_t nec_bitnum; // bit position
static uint8_t nec_code[4]; // decoded values


static void reset_state(void)
{
	// reset state
	nec_lastpin = _BV(IR_PIN); // save as idle, high
	nec_bitnum = 0;

	// idle timer1
	TIMSK &= ~_BV(OCIE0A); // disable the timer interrupt
	PRR |= _BV(PRTIM0); // power down timer 0
}


void ir_init(void)
{
	//PORTB |= _BV(IR_PIN); // pullup for IR pin (not necessary for some sensors)
	nec_lastpin = _BV(IR_PIN); // save as idle, high

	// configure timer0
#if TIMER0_PRESCALER == 64
	TCCR0B = (0<<CS02) | (1<<CS01) | (1<<CS00);
#else
	#error TODO: define prescaler
#endif
	PRR |= _BV(PRTIM0); // power down timer 0

	// configure pin change interrupt
	PCMSK |= _BV(IR_PIN); // configure pin change interrupt for IR pin
	GIFR = _BV(PCIF); // clear any possibly pending pin change interrupt
	GIMSK |= _BV(PCIE); // enable pin change interrupt
}


// pin change interrupt for edge sampling
ISR(PCINT0_vect)
{
	PROFILE_ENTER(PF_PCINT0_vect);
	uint8_t pin;
	pin = PINB & _BV(IR_PIN);
	uint8_t time;

	// toggle check
	if ((pin ^ nec_lastpin) == 0)
	{
		goto exit; // a glitch, our pin didn't change
	}
	
/* not necessary because of the toggle check
	if ((nec_bitnum & 0x01) ^ (pin == 1))
	{
		goto exit; // edge doesn't match bit number
	}
*/
	// if nec_bitnum==0, the timer is sleeping and we'll have to do it again
	time = TCNT0;
	OCR0A = time + MS2CLK(TIMEOUT); // refresh the timeout

	uint8_t diff;
	diff = time - nec_lasttime;

	if (nec_bitnum > 2) // normal case first, to optimize for it
	{
		if ((nec_bitnum & 0x01) == 0)
		{
			// even edges carry the information
			uint8_t bytenum;
			bytenum = (nec_bitnum - 4) >> 4;

			if (diff >= MS2CLK(TIME_LONG / TOLERANCE)
			 && diff <= MS2CLK(TIME_LONG * TOLERANCE))
			{
				nec_code[bytenum] >>= 1; // LSB first: shift right, add bits at top
				nec_code[bytenum] |= 0x80;
			}
			else if (diff >= MS2CLK(TIME_SHORT / TOLERANCE)
			      && diff <= MS2CLK(TIME_SHORT * TOLERANCE))
			{
				nec_code[bytenum] >>= 1; // LSB first: shift right, add bits at top
			}
			else
			{
				goto exit; // could be a glitch, don't change our state
			}
		}
		else
		{
			// for odd edge numbers, diff should be ~0.56 ms
			if (diff < MS2CLK(TIME_SHORT / TOLERANCE)
			 || diff > MS2CLK(TIME_SHORT * TOLERANCE))
			{
				goto exit; // could be a glitch, don't change our state
			}
		}

		if (nec_bitnum >= 67)
		{
			// done
			if ((nec_code[2] ^ nec_code[3]) == 0xFF) // command and complement
			{
				struct msg message; // message for result
				message.id = e_ir_cmd;
				message.data[0] = nec_code[0]; // address
				message.data[1] = nec_code[1]; // address complement or extension
				message.data[2] = nec_code[2]; // command
				msg_post(&message); // send it
			}

			reset_state();
			goto exit; // skip the default handling
		}
	}
	else if (nec_bitnum == 1) // AGC time
	{
		// diff should be ~9 ms
		if (diff < MS2CLK(TIME_AGC / TOLERANCE)
		 || diff > MS2CLK(TIME_AGC * TOLERANCE))
		{
			goto exit; // could be a glitch, don't change our state
		}
	}
	else if (nec_bitnum == 2) // space
	{
		// diff should be ~4.5 ms, or 2.25 ms for repeat code
		if (diff < MS2CLK(TIME_SPACE / TOLERANCE)
		 || diff > MS2CLK(TIME_SPACE * TOLERANCE))
		{
			goto exit; // could be a glitch, don't change our state
		}
		// TODO: repeat is not decoded atm
	}
	else // (nec_bitnum == 0)  first edge
	{
		PRR &= ~_BV(PRTIM0); // power up timer 0

		// setup timer 1 for timeout
		nec_lasttime = time = TCNT0; // re-do with awake timer
		OCR0A = time + MS2CLK(TIMEOUT); // re-do with awake timer
		TIFR |= _BV(OCF0A); // clear match interrups which got set while disabled
		TIMSK |= _BV(OCIE0A); // arm the timer interrupt

		nec_lastpin = 0;
		nec_bitnum = 1;
		goto exit;
	}

	nec_bitnum++;
	nec_lastpin = pin;
	nec_lasttime = time;
exit:
	PROFILE_EXIT(PF_PCINT0_vect);
}


// timeout interrupt
ISR(TIMER0_COMPA_vect)
{
	PROFILE_ENTER(TIMER0_COMPA_vect);

	reset_state();

	PROFILE_EXIT(TIMER0_COMPA_vect);
}
