#include <stdint.h>
#include <avr/io.h> // for device register definitions
#include <avr/interrupt.h> // for sei()
#include "config.h"
#include "msg.h"
#include "uart_oc.h"
#include "profiler.h"
#include "ir_nec.h"

// infrared address and codes
#if 1 // real target
#define IR_ADDR 0x78 // Yamaha Pianocraft
#define IR_STOP     0x01
#define IR_PLAY     0x02
#define IR_NEXT     0x03
#define IR_PREV     0x04
#define IR_VOL_UP   0x1E
#define IR_VOL_DOWN 0x1F
#else // for testing
#define IR_ADDR 0x6E // Squeezebox Touch
#define IR_STOP     0x04 // pause
#define IR_PLAY     0x08 // play
#define IR_NEXT     0x05 // fwd
#define IR_PREV     0x03 // rew
#define IR_VOL_UP   0x01
#define IR_VOL_DOWN 0x00
#endif

// Archos serial commands
#define ARCHOS_STOP     0xC2
#define ARCHOS_PLAY     0xC1
#define ARCHOS_NEXT     0xC8
#define ARCHOS_PREV     0xC4
#define ARCHOS_VOL_UP   0xD0
#define ARCHOS_VOL_DOWN 0xE0



static void handle_ir(uint8_t addr, uint8_t naddr, uint8_t cmd)
{
	uint8_t archoscmd = 0;
	
	if ((addr ^ naddr) != 0xFF)
	{
		return;
	}

	if (addr != IR_ADDR) // our device
	{
		return;
	}

	switch (cmd)
	{
	case IR_STOP: // stop
		archoscmd  = ARCHOS_STOP;
		break;

	case IR_PLAY: // play/pause
		archoscmd  = ARCHOS_PLAY;
		break;

	case IR_NEXT: // next
		archoscmd  = ARCHOS_NEXT;
		break;

	case IR_PREV: // prev
		archoscmd  = ARCHOS_PREV;
		break;

	case IR_VOL_UP: // Vol+
		//archoscmd  = ARCHOS_VOL_UP;
		break;

	case IR_VOL_DOWN: // Vol-
		//archoscmd  = ARCHOS_VOL_DOWN;
		break;

	default:
		break;
	}

	if (archoscmd)
	{
		//USI_UART_Transmit_Byte(archoscmd);
		uart_byte_send(archoscmd);
	}
}


int main(void)
{
	ACSR |= _BV(ACD); // disable analog comparator, saves bandgap current
	PRR = _BV(PRADC) | _BV(PRUSI); // power down unused components
	DIDR0 = 0x3F & ~_BV(CFG_PIN_IR); // disable all input buffers of pins not being read
	// in power-down mode, this brings us down to 200 nA (with DebugWire and BOD off)
	
	profiler_init();
	uart_init();
	ir_init();

	sei(); // global interrupt enable

	while (1)
	{
		struct msg message;

		message = msg_get();
		switch (message.id)
		{
		case e_ir_cmd: // infrared command received, wparam is code
			handle_ir(message.data[0], message.data[1], message.data[2]);
			break;

		default: // unknown message, should not happen
			break;
		}
	}

	return 0;
}
