/*! \file msg.c
    \brief Interrupt-safe message queue implementation
    
    A mean and lean implementation of a message queue.
    Main thread or interrupts can post messages,
    the foreground loop waits and picks a message.
    The controller is sent to a low power sleep mode while waiting.
*/

#include <stdint.h>
#include <avr/sleep.h> // for power-save idle sleep
#include <avr/interrupt.h> // for cli/sei
#include "msg.h"


struct msg msg_queue[MSG_QUEUESIZE];
volatile uint8_t msg_head;
volatile uint8_t msg_tail;


/*************** private functions ***************/

// find out if we can go to deep sleep, system dependent!!!
static inline uint8_t peripherals_active(void)
{
	// for this design, timer 0 is the only one needing CLKio

	// return true if any of the power save candidates is active
	return ((PRR & (_BV(PRTIM0) | _BV(PRTIM1) | _BV(PRUSI) | _BV(PRADC)))
	            != (_BV(PRTIM0) | _BV(PRTIM1) | _BV(PRUSI) | _BV(PRADC)));
}



/*************** public API functions ***************/

// return the fill level, how many messages are waiting
uint8_t msg_fill(void)
{
	return ((msg_head - msg_tail) & MSG_MASK);
}


// post a message to the head, *not* allowed from interrupt context
void msg_post_front(struct msg* message)
{
	msg_queue[msg_tail] = *message; // store received message in queue
	msg_tail = (msg_tail - 1) & MSG_MASK; // store new index
}


// wait for a message
struct msg msg_get(void)
{
	uint8_t tmptail;

    for(;;) 
    {
        uint8_t sreg;

		sreg = SREG;
        cli(); // make the check below atomic, else in rare race condition we may sleep despite new msg.
        if (msg_head != msg_tail)
        {
            SREG = sreg; // sei();
            break;
        }
		
		// use deep power down sleep if peripherals are not required
		set_sleep_mode(peripherals_active() ? SLEEP_MODE_IDLE : SLEEP_MODE_PWR_DOWN);
		
		sleep_enable();
        SREG = sreg; // sei();
	    sleep_cpu(); // trick: the instruction behind sei gets executed atomic, too
	    sleep_disable();
    }

	tmptail = (msg_tail + 1) & MSG_MASK; // calculate buffer index
	
	msg_tail = tmptail; // store new index
	return msg_queue[tmptail]; // return message
}
