// very simple Tx only UART, using timer compare unit for bit timing
// special: also have an inverted output, with a lead time for falling edge

#include <stdint.h>
#include <avr/io.h> // for device register definitions
#include <avr/interrupt.h> // for interrupt handlers
#include "config.h"
#include "profiler.h"
#include "uart_oc.h" // own interface


#define BAUDRATE 9600

#define TIMER1_PRESCALER 1

#ifdef CFG_PIN_TX
#define TX_PIN CFG_PIN_TX // pin definition
#else
#error Missing CFG_PIN_TX definition in config.h
#endif

#ifdef CFG_PIN_POWER
#define NTX_PIN CFG_PIN_POWER // pin definition
#else
#error Missing CFG_PIN_POWER definition in config.h
#endif


#if TIMER1_PRESCALER == 1
#define TIMER_SPEED ((0<<CS13) | (0<<CS12) | (0<<CS11) | (1<<CS10))
#else
#error Implement for TIMER1_PRESCALER
#endif


#define MS2CLK(ms) ((uint8_t)((ms)*F_CPU/TIMER1_PRESCALER/1000))

// timing advance for the power pin, in milliseconds
// note: must not be larger than bit time minus ISR time! (0.1 ms)
#define TIME_LEAD 0.005 

// timer reload value for selected baudrate
#define BR_RELOAD ((F_CPU / TIMER1_PRESCALER / BAUDRATE) - 1)



static uint8_t uart_byte; // data to send out
static uint8_t uart_bitnum; // counting the outgoing bits


void uart_init(void)
{
	// configure pins
	PORTB |= _BV(TX_PIN); // drive Tx line idle high
	PORTB &= ~_BV(NTX_PIN); // drive /Tx line idle low
	DDRB |= _BV(TX_PIN) | _BV(NTX_PIN); // outputs

	// fine-tune OSCCAL to better match baudrate
	// TODO: this is individually chip-dependent!!!
	OSCCAL += 2;

	// configure timer1 for baudrate generation
	OCR1C = OCR1A = BR_RELOAD;

	TCCR1 = _BV(CTC1); // clear on OCR1C match

	// init compare outputs
	TCCR1 = (TCCR1 & ~(_BV(COM1A1) | _BV(COM1A0)))
	 | (1<<COM1A1) | (1<<COM1A0); // set OC1A on next match

	GTCCR = (GTCCR & ~(_BV(COM1B1) | _BV(COM1B0)))
		| (1<<COM1B1) | (0<<COM1B0) // clear OC1B on next match
		| _BV(FOC1B) | _BV(FOC1A); // force OC lines to desired idle

	PRR |= _BV(PRTIM1); // power down timer 1
}


// very primitive single byte sending, assumes Tx is idle
void uart_byte_send(uint8_t byte)
{
	if (uart_bitnum != 0)
	{
		return; // error, Tx is busy
	}
	
	uart_byte = byte;
	uart_bitnum = 1;

	PRR &= ~_BV(PRTIM1); // power up timer 1

	TIFR = _BV(OCF1A); // clear any potentially pending interrupt
	TIMSK |= _BV(OCIE1A); // enable interrupt

	TCNT1 = BR_RELOAD - MS2CLK(TIME_LEAD) - 1; // edge action asap

	// prepare start bit
	OCR1B = BR_RELOAD - MS2CLK(TIME_LEAD);

	GTCCR = (GTCCR & ~(_BV(COM1B1) | _BV(COM1B0)))
		| (1<<COM1B1) | (1<<COM1B0); // set OC1B on next match (start bit)

	// start the timer
	TCCR1 = (TCCR1 & ~( _BV(CS13) | _BV(CS12) | _BV(CS11) | _BV(CS10)
	                  | _BV(COM1A1) | _BV(COM1A0)))
		| TIMER_SPEED
		| (1<<COM1A1) | (0<<COM1A0); // clear OC1A on next match (start bit)
}


// timeout interrupt
ISR(TIMER1_COMPA_vect)
{
	PROFILE_ENTER(TIMER1_COMPA_vect);

	if (uart_bitnum <= 8) // data bits
	{
		if (uart_byte & 0x01) // LSB first
		{
			OCR1B = BR_RELOAD; // simultaneous

			TCCR1 = (TCCR1 & ~(_BV(COM1A1) | _BV(COM1A0)))
			 | (1<<COM1A1) | (1<<COM1A0); // set OC1A on next match

			GTCCR = (GTCCR & ~(_BV(COM1B1) | _BV(COM1B0)))
				| (1<<COM1B1) | (0<<COM1B0); // clear OC1B on next match
		}
		else
		{
			OCR1B = BR_RELOAD - MS2CLK(TIME_LEAD); // falling edge gets lead time

			TCCR1 = (TCCR1 & ~(_BV(COM1A1) | _BV(COM1A0)))
			 | (1<<COM1A1) | (0<<COM1A0); // clear OC1A on next match

			GTCCR = (GTCCR & ~(_BV(COM1B1) | _BV(COM1B0)))
				| (1<<COM1B1) | (1<<COM1B0); // set OC1B on next match
		}
		uart_byte >>= 1;
	}
	else if (uart_bitnum == 9) // stop bit
	{
		OCR1B = BR_RELOAD; // simultaneous

		TCCR1 = (TCCR1 & ~(_BV(COM1A1) | _BV(COM1A0)))
		 | (1<<COM1A1) | (1<<COM1A0); // set OC1A on next match

		GTCCR = (GTCCR & ~(_BV(COM1B1) | _BV(COM1B0)))
			| (1<<COM1B1) | (0<<COM1B0); // clear OC1B on next match
	}
	else // end of byte
	{
		TIMSK &= ~_BV(OCIE1A); // disable interrupt

		// stop the timer
		TCCR1 &= ~(_BV(CS13) | _BV(CS12) | _BV(CS11) | _BV(CS10));
		PRR |= _BV(PRTIM1); // power down timer 1

		uart_bitnum = 0; // idle
	}
	
	if (uart_bitnum) // still active?
	{
		uart_bitnum++;
	}

	PROFILE_EXIT(TIMER1_COMPA_vect);
}
